from yt.utilities.exceptions import YTModuleRemoved

raise YTModuleRemoved(
    "halo_analysis",
    "https://github.com/yt-project/yt_astro_analysis",
    "https://yt-astro-analysis.readthedocs.io/",
)
